@@ -24,6 +24,8 @@
 #ifndef BONOBO_ACTIVATION_ID_H
 #define BONOBO_ACTIVATION_ID_H
 
+#ifndef BONOBO_DISABLE_DEPRECATED
+
 #include <bonobo-activation/Bonobo_Activation_types.h>
 
 G_BEGIN_DECLS
@@ -48,4 +50,6 @@ void                   bonobo_activation_info_free           (BonoboActivationIn
 
 G_END_DECLS
 
+#endif /* BONOBO_DISABLE_DEPRECATED */
+
 #endif /* BONOBO_ACTIVATION_ID_H */
