@@ -3,7 +3,7 @@
  *  oafd: OAF CORBA dameon.
  *
  *  Copyright (C) 1999, 2000 Red Hat, Inc.
- *  Copyright (C) 1999, 2000 Eazel, Inc.
+ *  Copyrigfht (C) 1999, 2000 Eazel, Inc.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License as
@@ -37,32 +37,50 @@
 #include "bonobo-activation/bonobo-activation-i18n.h"
 #include "activation-server-corba-extensions.h"
 
+/*
+ *    Time delay after all servers are de-registered / dead
+ * before quitting the server.
+ */
+#define IDLE_QUIT_TIMEOUT 1000
+
+/*
+ *    We always have the local NamingContext in the
+ * registry at exit, so quit if only 1 left.
+ */
+#define RESIDUAL_SERVERS 1
+
 /*** App-specific servant structures ***/
 
 typedef struct {
+        /* CORBA servant structure */
 	POA_Bonobo_ObjectDirectory servant;
-	PortableServer_POA poa;
+
+        /* Information on all servers */
+	GHashTable           *by_iid;
 	Bonobo_ServerInfoList attr_servers;
-	Bonobo_CacheTime time_list_changed;
+	Bonobo_CacheTime      time_list_changed;
 
-	GHashTable *by_iid;
+	CORBA_char           *attr_domain;
 
-	CORBA_char *attr_domain;
-	const char *attr_hostID;
-
-	guchar is_locked;
-
-	GHashTable *active_servers;
+        /* CORBA Object tracking */
+	GHashTable      *active_servers;
+        guint            no_servers_timeout;
 	Bonobo_CacheTime time_active_changed;
 
-        char **registry_source_directories;
-        time_t time_did_stat;
-        GHashTable *registry_directory_mtimes;
+        /* Source polling bits */
+        char           **registry_source_directories;
+        time_t           time_did_stat;
+        GHashTable      *registry_directory_mtimes;
 
 	CORBA_Object self;
 } impl_POA_Bonobo_ObjectDirectory;
 
-/*** Stub implementations ***/
+/*
+ * DON'T FIXME: this smells, here we vandalise the oh so
+ * complicated design and enforce the invariant that there
+ * can only ever be one ObjectDirectory in-proc.
+ */
+static impl_POA_Bonobo_ObjectDirectory *main_dir = NULL;
 
 #ifdef BONOBO_ACTIVATION_DEBUG
 static void
@@ -77,7 +95,7 @@ od_dump_list (impl_POA_Bonobo_ObjectDirectory * od)
 			 od->attr_servers._buffer[i].location_info);
 		for (j = 0; j < od->attr_servers._buffer[i].props._length;
 		     j++) {
-			Bonobo_Property *prop =
+			Bonobo_ActivationProperty *prop =
 				&(od->attr_servers._buffer[i].
 				  props._buffer[j]);
                         if (strchr (prop->name, '-') != NULL) /* Translated, likely to
@@ -86,18 +104,18 @@ od_dump_list (impl_POA_Bonobo_ObjectDirectory * od)
 
 			g_print ("    %s = ", prop->name);
 			switch (prop->v._d) {
-			case Bonobo_P_STRING:
+			case Bonobo_ACTIVATION_P_STRING:
 				g_print ("\"%s\"\n", prop->v._u.value_string);
 				break;
-			case Bonobo_P_NUMBER:
+			case Bonobo_ACTIVATION_P_NUMBER:
 				g_print ("%f\n", prop->v._u.value_number);
 				break;
-			case Bonobo_P_BOOLEAN:
+			case Bonobo_ACTIVATION_P_BOOLEAN:
 				g_print ("%s\n",
 					 prop->v.
 					 _u.value_boolean ? "TRUE" : "FALSE");
 				break;
-			case Bonobo_P_STRINGV:
+			case Bonobo_ACTIVATION_P_STRINGV:
 				g_print ("[");
 				for (k = 0;
 				     k < prop->v._u.value_stringv._length;
@@ -120,7 +138,7 @@ od_dump_list (impl_POA_Bonobo_ObjectDirectory * od)
 
 static gboolean
 registry_directory_needs_update (impl_POA_Bonobo_ObjectDirectory *servant,
-                                 const char *directory)
+                                 const char                      *directory)
 {
         gboolean needs_update;
         struct stat statbuf;
@@ -130,8 +148,8 @@ registry_directory_needs_update (impl_POA_Bonobo_ObjectDirectory *servant,
                 return FALSE;
         }
  
-        old_mtime = (time_t) g_hash_table_lookup (servant->registry_directory_mtimes,
-                                                  directory);
+        old_mtime = (time_t) g_hash_table_lookup (
+                servant->registry_directory_mtimes, directory);
 
         g_hash_table_insert (servant->registry_directory_mtimes,
                              (gpointer) directory,
@@ -157,20 +175,24 @@ update_registry (impl_POA_Bonobo_ObjectDirectory *servant, gboolean force_reload
         gboolean must_load;
         static int reload_recurse_depth = 0;
 
-        /* FIXME bugzilla.eazel.com 2727: we should only reload those directories that have
-         * actually changed instead of reloading all when any has
-         * changed. 
-         */
+        
+
+#ifdef BONOBO_ACTIVATION_DEBUG
+        g_warning ("Update registry %d", reload_recurse_depth);
+#endif
 
         if (reload_recurse_depth++)
                 return;
 
         must_load = FALSE;
         
-        /* Don't stat more than once every 5 seconds or activation
-           could be too slow. This works even on the first read
-           because then `time_did_stat is 0' */
+        /*
+         * Don't stat more than once every 5 seconds or activation
+         * could be too slow. This works even on the first read
+         * because then `time_did_stat is 0'
+         */
         cur_time = time (NULL);
+
         if (cur_time - 5 > servant->time_did_stat) {
                 servant->time_did_stat = cur_time;
                 
@@ -183,18 +205,26 @@ update_registry (impl_POA_Bonobo_ObjectDirectory *servant, gboolean force_reload
         }
         
         if (must_load || force_reload) {
-                Bonobo_ServerInfo_load (servant->registry_source_directories,
-                                        &servant->attr_servers,
-                                        &servant->by_iid,
-                                        servant->attr_hostID,
-                                        servant->attr_domain);
-                servant->time_list_changed = time (NULL);
+                /*
+                 * FIXME bugzilla.eazel.com 2727: we should only reload those
+                 * directories that have actually changed instead of reloading
+                 * all when any has changed. 
+                 */
+#ifdef BONOBO_ACTIVATION_DEBUG
+                g_warning ("Re-load %d %d", must_load, force_reload);
+#endif
+                bonobo_server_info_load (servant->registry_source_directories,
+                                         &servant->attr_servers,
+                                         &servant->by_iid,
+                                         bonobo_activation_hostname_get (),
+                                         servant->attr_domain);
+                servant->time_did_stat = servant->time_list_changed = time (NULL);
 
 #ifdef BONOBO_ACTIVATION_DEBUG
                 od_dump_list (servant);
 #endif
                 if (must_load)
-                        notify_clients_cache_reset ();
+                        activation_clients_cache_notify ();
         }
 
         reload_recurse_depth--;
@@ -233,9 +263,8 @@ split_path_unique (const char *colon_delimited_path)
 
         ret = g_new (char *, max + 1);
 
-        for (l = tmp, i = 0; l; l = l->next) {
+        for (l = tmp, i = 0; l; l = l->next)
                 ret [i++] = l->data;
-        }
 
         ret [i] = NULL;
 
@@ -246,9 +275,10 @@ split_path_unique (const char *colon_delimited_path)
 }
 
 static Bonobo_ServerInfoListCache *
-impl_Bonobo_ObjectDirectory__get_servers (impl_POA_Bonobo_ObjectDirectory * servant,
-				       Bonobo_CacheTime only_if_newer,
-				       CORBA_Environment * ev)
+impl_Bonobo_ObjectDirectory__get_servers (
+        impl_POA_Bonobo_ObjectDirectory *servant,
+        Bonobo_CacheTime                 only_if_newer,
+        CORBA_Environment               *ev)
 {
 	Bonobo_ServerInfoListCache *retval;
 
@@ -266,24 +296,22 @@ impl_Bonobo_ObjectDirectory__get_servers (impl_POA_Bonobo_ObjectDirectory * serv
 	return retval;
 }
 
-typedef struct
-{
+typedef struct {
 	Bonobo_ImplementationID *seq;
-	int last_used;
-}
-StateCollectionInfo;
+	int                      last_used;
+} StateCollectionInfo;
 
 static void
-add_active_server (char *key, gpointer value, StateCollectionInfo * sci)
+collate_active_server (char *key, gpointer value, StateCollectionInfo *sci)
 {
-	sci->seq[(sci->last_used)++] = CORBA_string_dup (key);
+	sci->seq [(sci->last_used)++] = CORBA_string_dup (key);
 }
 
 static Bonobo_ServerStateCache *
-impl_Bonobo_ObjectDirectory_get_active_servers (impl_POA_Bonobo_ObjectDirectory *
-					     servant,
-					     Bonobo_CacheTime only_if_newer,
-					     CORBA_Environment * ev)
+impl_Bonobo_ObjectDirectory_get_active_servers (
+        impl_POA_Bonobo_ObjectDirectory *servant,
+        Bonobo_CacheTime                 only_if_newer,
+        CORBA_Environment               *ev)
 {
 	Bonobo_ServerStateCache *retval;
 
@@ -301,7 +329,7 @@ impl_Bonobo_ObjectDirectory_get_active_servers (impl_POA_Bonobo_ObjectDirectory
 		sci.last_used = 0;
 
 		g_hash_table_foreach (servant->active_servers,
-				      (GHFunc) add_active_server, &sci);
+				      (GHFunc) collate_active_server, &sci);
 		CORBA_sequence_set_release (&(retval->_u.active_servers),
 					    CORBA_TRUE);
 	}
@@ -309,29 +337,6 @@ impl_Bonobo_ObjectDirectory_get_active_servers (impl_POA_Bonobo_ObjectDirectory
 	return retval;
 }
 
-static CORBA_char *
-impl_Bonobo_ObjectDirectory__get_domain (impl_POA_Bonobo_ObjectDirectory * servant,
-				      CORBA_Environment * ev)
-{
-	return CORBA_string_dup (servant->attr_domain);
-}
-
-static CORBA_char *
-impl_Bonobo_ObjectDirectory__get_hostID (impl_POA_Bonobo_ObjectDirectory * servant,
-				      CORBA_Environment * ev)
-{
-	return CORBA_string_dup (servant->attr_hostID);
-}
-
-static CORBA_char *
-impl_Bonobo_ObjectDirectory__get_username (impl_POA_Bonobo_ObjectDirectory *
-					servant, CORBA_Environment * ev)
-{
-	return CORBA_string_dup (g_get_user_name ());
-}
-
-
-
 static CORBA_Object 
 od_get_active_server (impl_POA_Bonobo_ObjectDirectory *servant,
                       Bonobo_ImplementationID          iid,
@@ -342,7 +347,8 @@ od_get_active_server (impl_POA_Bonobo_ObjectDirectory *servant,
         char *display;
         char *display_iid;
 
-        display = activation_server_CORBA_Context_get_value (ctx, "display", NULL, ev);
+        display = activation_server_CORBA_Context_get_value (
+                ctx, "display", NULL, ev);
         
         if (display != NULL) {
                 display_iid = g_strconcat (display, ",", iid, NULL);
@@ -353,28 +359,27 @@ od_get_active_server (impl_POA_Bonobo_ObjectDirectory *servant,
                 g_free (display_iid);
                 
                 if (retval != CORBA_OBJECT_NIL &&
-                    !CORBA_Object_non_existent (retval, ev)) {
+                    !CORBA_Object_non_existent (retval, ev))
                         return CORBA_Object_duplicate (retval, ev);
-                }
         }
 
         retval = g_hash_table_lookup (servant->active_servers, iid);
         
         if (retval != CORBA_OBJECT_NIL &&
-            !CORBA_Object_non_existent (retval, ev)) {
+            !CORBA_Object_non_existent (retval, ev))
                 return CORBA_Object_duplicate (retval, ev);
-        }
 
         return CORBA_OBJECT_NIL;
 }
 
 static CORBA_Object
-impl_Bonobo_ObjectDirectory_activate (impl_POA_Bonobo_ObjectDirectory *servant,
-                                      Bonobo_ImplementationID          iid,
-                                      Bonobo_ActivationContext         ac,
-                                      Bonobo_ActivationFlags           flags,
-                                      CORBA_Context                    ctx,
-                                      CORBA_Environment               *ev)
+impl_Bonobo_ObjectDirectory_activate (
+        impl_POA_Bonobo_ObjectDirectory *servant,
+        Bonobo_ImplementationID          iid,
+        Bonobo_ActivationContext         ac,
+        Bonobo_ActivationFlags           flags,
+        CORBA_Context                    ctx,
+        CORBA_Environment               *ev)
 {
 	CORBA_Object retval;
 	Bonobo_ServerInfo *si;
@@ -391,9 +396,8 @@ impl_Bonobo_ObjectDirectory_activate (impl_POA_Bonobo_ObjectDirectory *servant,
         if (!(flags & Bonobo_ACTIVATION_FLAG_PRIVATE)) {
                 retval = od_get_active_server (servant, iid, ctx, ev);
 
-                if (retval != CORBA_OBJECT_NIL) {
+                if (retval != CORBA_OBJECT_NIL)
                         return retval;
-                }
         }
 
 	if (flags & Bonobo_ACTIVATION_FLAG_EXISTING_ONLY) {
@@ -442,9 +446,8 @@ impl_Bonobo_ObjectDirectory_activate (impl_POA_Bonobo_ObjectDirectory *servant,
 
                         CORBA_exception_free (&retry_ev);
                         
-                        if (retval != CORBA_OBJECT_NIL) {
+                        if (retval != CORBA_OBJECT_NIL)
                                 CORBA_exception_free (ev);
-                        }
                 }
         }
 
@@ -461,30 +464,133 @@ impl_Bonobo_ObjectDirectory_activate (impl_POA_Bonobo_ObjectDirectory *servant,
 	return retval;
 }
 
-static void
-impl_Bonobo_ObjectDirectory_lock (impl_POA_Bonobo_ObjectDirectory * servant,
-			       CORBA_Environment * ev)
-{
-	while (servant->is_locked)
-		g_main_context_iteration (NULL, TRUE);
+extern GMainLoop *main_loop;
 
-	servant->is_locked = TRUE;
+static gboolean
+quit_server_timeout (gpointer user_data)
+{
+#ifdef BONOBO_ACTIVATION_DEBUG
+        g_warning ("Quit server !");
+#endif
+
+        if (!main_dir ||
+            g_hash_table_size (main_dir->active_servers) > RESIDUAL_SERVERS)
+                g_warning ("Serious error hanlding server count, quitting anyway");
+
+        g_main_loop_quit (main_loop);
+
+        return FALSE;
 }
 
 static void
-impl_Bonobo_ObjectDirectory_unlock (impl_POA_Bonobo_ObjectDirectory * servant,
-				 CORBA_Environment * ev)
+check_quit (impl_POA_Bonobo_ObjectDirectory *servant)
 {
-	g_return_if_fail (servant->is_locked);
+        if (g_hash_table_size (servant->active_servers) > RESIDUAL_SERVERS) {
+                if (servant->no_servers_timeout != 0)
+                        g_source_remove (servant->no_servers_timeout);
+                servant->no_servers_timeout = 0;
+        } else {
+                
+                servant->no_servers_timeout = g_timeout_add (
+                        IDLE_QUIT_TIMEOUT, quit_server_timeout, NULL);
+        }
 
-	servant->is_locked = FALSE;
+	servant->time_active_changed = time (NULL);
+}
+
+static gboolean
+prune_dead_servers (gpointer key,
+                    gpointer value,
+                    gpointer user_data)
+{
+        gboolean dead;
+        ORBitConnectionStatus status;
+
+        status = ORBit_small_get_connection_status (value);
+
+        dead = (status == ORBIT_CONNECTION_DISCONNECTED);
+
+#ifdef BONOBO_ACTIVATION_DEBUG
+        fprintf (stderr, "IID '%20s', %s\n", (char *) key,
+                 dead ? "dead" : "alive");
+#endif
+        
+        return dead;
+}
+
+static void
+active_server_cnx_broken (ORBitConnection *cnx,
+                          gpointer         dummy)
+
+{
+        impl_POA_Bonobo_ObjectDirectory *servant = main_dir;
+
+        if (!servant) /* shutting down */
+                return;
+
+        g_hash_table_foreach_remove (servant->active_servers,
+                                     prune_dead_servers, servant);
+#ifdef BONOBO_ACTIVATION_DEBUG
+        g_warning ("After prune: %d live servers",
+                   g_hash_table_size (
+                           servant->active_servers) - RESIDUAL_SERVERS);
+#endif
+
+        activation_clients_check ();
+
+        check_quit (servant);
+}
+
+static void
+add_active_server (impl_POA_Bonobo_ObjectDirectory *servant,
+                   const char                      *iid,
+                   CORBA_Object                     object)
+{
+        ORBitConnection *cnx;
+
+        cnx = ORBit_small_get_connection (object);
+        if (cnx) {
+                if (!g_object_get_data (G_OBJECT (cnx), "object_count")) {
+                        g_object_set_data (
+                                G_OBJECT (cnx), "object_count", GUINT_TO_POINTER (1));
+                        
+                        g_signal_connect (
+                                cnx, "broken",
+                                G_CALLBACK (active_server_cnx_broken),
+                                NULL);
+                }
+        } else
+                g_assert (!strcmp (iid, NAMING_CONTEXT_IID));
+
+	g_hash_table_replace (servant->active_servers,
+                              g_strdup (iid),
+                              CORBA_Object_duplicate (object, NULL));
+
+        check_quit (servant);
+}
+
+static void
+active_server_free (gpointer active_server)
+{
+        CORBA_Object_release (active_server, NULL);
+}
+
+static void
+remove_active_server (impl_POA_Bonobo_ObjectDirectory *servant,
+                      const char                      *iid,
+                      CORBA_Object                     object)
+{
+	g_hash_table_remove (servant->active_servers, iid);
+
+        check_quit (servant);
 }
 
 static Bonobo_RegistrationResult
-impl_Bonobo_ObjectDirectory_register_new (impl_POA_Bonobo_ObjectDirectory *servant,
-                                          Bonobo_ImplementationID          iid,
-                                          CORBA_Object                     obj,
-                                          CORBA_Environment               *ev)
+impl_Bonobo_ObjectDirectory_register_new (
+        impl_POA_Bonobo_ObjectDirectory *servant,
+        Bonobo_ImplementationID          iid,
+        CORBA_Object                     obj,
+        CORBA_Environment               *ev)
 {
 	CORBA_Object oldobj;
         Bonobo_ImplementationID actual_iid;
@@ -506,40 +612,59 @@ impl_Bonobo_ObjectDirectory_register_new (impl_POA_Bonobo_ObjectDirectory *serva
 	if (!g_hash_table_lookup (servant->by_iid, actual_iid))
 		return Bonobo_ACTIVATION_REG_NOT_LISTED;
 
-        CORBA_Object_release (oldobj, ev);
 #ifdef BONOBO_ACTIVATION_DEBUG
         g_warning ("Server register. '%s' : %p", iid, obj);
 #endif
-	g_hash_table_insert (servant->active_servers,
-			     oldobj ? iid : g_strdup (iid),
-			     CORBA_Object_duplicate (obj, ev));
-	servant->time_active_changed = time (NULL);
+
+        add_active_server (servant, iid, obj);
 
 	return Bonobo_ACTIVATION_REG_SUCCESS;
 }
 
 static void
-impl_Bonobo_ObjectDirectory_unregister (impl_POA_Bonobo_ObjectDirectory * servant,
-                                        Bonobo_ImplementationID iid,
-                                        CORBA_Object obj,
-                                        Bonobo_ObjectDirectory_UnregisterType notify,
-                                        CORBA_Environment * ev)
+impl_Bonobo_ObjectDirectory_unregister (
+        impl_POA_Bonobo_ObjectDirectory      *servant,
+        Bonobo_ImplementationID               iid,
+        CORBA_Object                          obj,
+        CORBA_Environment                    *ev)
 {
-	char *orig_iid;
-	CORBA_Object orig_obj;
+	CORBA_Object orig_object;
 
-	if (!g_hash_table_lookup_extended
-	    (servant->active_servers, iid, (gpointer *) & orig_iid,
-	     (gpointer *) & orig_obj)
-	    || !CORBA_Object_is_equivalent (orig_obj, obj, ev))
-		return;
+        orig_object = g_hash_table_lookup (servant->active_servers, iid);
+        
+        if (orig_object == CORBA_OBJECT_NIL ||
+	    !CORBA_Object_is_equivalent (orig_object, obj, ev)) {
+                CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+                                     ex_Bonobo_ObjectDirectory_NotRegistered,
+                                     NULL);
+                return;
+        }
 
-	g_hash_table_remove (servant->active_servers, iid);
+        remove_active_server (servant, iid, obj);
+}
 
-	g_free (orig_iid);
-	CORBA_Object_release (orig_obj, ev);
+static CORBA_char *
+impl_Bonobo_ObjectDirectory__get_domain (
+        impl_POA_Bonobo_ObjectDirectory *servant,
+        CORBA_Environment               *ev)
+{
+	return CORBA_string_dup (servant->attr_domain);
+}
 
-	servant->time_active_changed = time (NULL);
+static CORBA_char *
+impl_Bonobo_ObjectDirectory__get_hostID (
+        impl_POA_Bonobo_ObjectDirectory *servant,
+        CORBA_Environment               *ev)
+{
+        return CORBA_string_dup (bonobo_activation_hostname_get ());
+}
+
+static CORBA_char *
+impl_Bonobo_ObjectDirectory__get_username (
+        impl_POA_Bonobo_ObjectDirectory *servant,
+        CORBA_Environment               *ev)
+{
+	return CORBA_string_dup (g_get_user_name ());
 }
 
 /*** epv structures ***/
@@ -566,8 +691,6 @@ static POA_Bonobo_ObjectDirectory__epv impl_Bonobo_ObjectDirectory_epv = {
 	(gpointer) &impl_Bonobo_ObjectDirectory__get_hostID,
 	(gpointer) &impl_Bonobo_ObjectDirectory__get_domain,
 	(gpointer) &impl_Bonobo_ObjectDirectory_activate,
-	(gpointer) &impl_Bonobo_ObjectDirectory_lock,
-	(gpointer) &impl_Bonobo_ObjectDirectory_unlock,
 	(gpointer) &impl_Bonobo_ObjectDirectory_register_new,
 	(gpointer) &impl_Bonobo_ObjectDirectory_unregister
 };
@@ -580,12 +703,74 @@ static POA_Bonobo_ObjectDirectory__vepv impl_Bonobo_ObjectDirectory_vepv = {
 	&impl_Bonobo_ObjectDirectory_epv
 };
 
-/*
- * FIXME: this smells, here we vandalise the oh so complicated
- * design and enforce the invariant that there can only ever
- * be one ObjectDirectory.
- */
-static impl_POA_Bonobo_ObjectDirectory *main_dir = NULL;
+Bonobo_ObjectDirectory
+bonobo_object_directory_get (void)
+{
+        if (!main_dir)
+                return CORBA_OBJECT_NIL;
+        else
+                return main_dir->self;
+}
+
+void
+bonobo_object_directory_init (PortableServer_POA poa,
+                              const char        *domain,
+                              const char        *registry_path,
+                              CORBA_Environment *ev)
+{
+	PortableServer_ObjectId         *objid;
+	impl_POA_Bonobo_ObjectDirectory *servant;
+
+        g_assert (main_dir == NULL);
+
+	servant = g_new0 (impl_POA_Bonobo_ObjectDirectory, 1);
+
+        main_dir = servant; /* DON'T FIXME: this is good. */
+
+	servant->servant.vepv = &impl_Bonobo_ObjectDirectory_vepv;
+	POA_Bonobo_ObjectDirectory__init ((PortableServer_Servant) servant,
+				       ev);
+	objid = PortableServer_POA_activate_object (poa, servant, ev);
+	CORBA_free (objid);
+	servant->self =
+		PortableServer_POA_servant_to_reference (poa, servant, ev);
+
+	servant->attr_domain = g_strdup (domain);
+	servant->by_iid = NULL;
+
+        servant->registry_source_directories = split_path_unique (registry_path);
+        servant->registry_directory_mtimes = g_hash_table_new (g_str_hash, g_str_equal);
+
+        update_registry (servant, FALSE);
+
+        servant->active_servers =
+                g_hash_table_new_full (g_str_hash, g_str_equal,
+                                       g_free, active_server_free);
+        servant->no_servers_timeout = 0;
+}
+
+void
+bonobo_object_directory_shutdown (PortableServer_POA poa,
+                                  CORBA_Environment *ev)
+{
+	PortableServer_ObjectId *oid;
+	impl_POA_Bonobo_ObjectDirectory *servant = main_dir;
+
+	oid = PortableServer_POA_servant_to_id (poa, servant, ev);
+	PortableServer_POA_deactivate_object (poa, oid, ev);
+	CORBA_free (oid);
+
+        main_dir = NULL;
+
+        CORBA_Object_release (servant->self, ev);
+        g_hash_table_destroy (servant->active_servers);
+        servant->active_servers = NULL;
+        g_hash_table_destroy (servant->registry_directory_mtimes);
+        servant->registry_directory_mtimes = NULL;
+        g_strfreev (servant->registry_source_directories);
+
+        g_free (servant);
+}
 
 CORBA_Object
 bonobo_object_directory_re_check_fn (const char        *display,
@@ -625,48 +810,8 @@ bonobo_object_directory_re_check_fn (const char        *display,
         return retval;
 }
 
-Bonobo_ObjectDirectory
-Bonobo_ObjectDirectory_create (PortableServer_POA poa,
-                               const char        *domain,
-                               const char        *registry_path,
-                               CORBA_Environment *ev)
-{
-	Bonobo_ObjectDirectory retval;
-	impl_POA_Bonobo_ObjectDirectory *newservant;
-	PortableServer_ObjectId *objid;
-
-        g_assert (main_dir == NULL);
-
-	newservant = g_new0 (impl_POA_Bonobo_ObjectDirectory, 1);
-
-        main_dir = newservant;
-
-	newservant->servant.vepv = &impl_Bonobo_ObjectDirectory_vepv;
-	newservant->poa = poa;
-	POA_Bonobo_ObjectDirectory__init ((PortableServer_Servant) newservant,
-				       ev);
-	objid = PortableServer_POA_activate_object (poa, newservant, ev);
-	CORBA_free (objid);
-	newservant->self = retval =
-		PortableServer_POA_servant_to_reference (poa, newservant, ev);
-
-	newservant->attr_domain = g_strdup (domain);
-	newservant->attr_hostID = bonobo_activation_hostname_get ();
-	newservant->by_iid = NULL;
-
-        newservant->registry_source_directories = split_path_unique (registry_path);
-        newservant->registry_directory_mtimes = g_hash_table_new (g_str_hash, g_str_equal);
-
-        update_registry (newservant, FALSE);
-
-        newservant->active_servers =
-                g_hash_table_new (g_str_hash, g_str_equal);
-
-	return CORBA_Object_duplicate (retval, ev);
-}
-
 void
-reload_object_directory (void)
+bonobo_object_directory_reload (void)
 {
         g_print ("reloading our object directory!\n");
         update_registry (main_dir, TRUE);
